<?php

class Condition
{
  var $type;     // '=', '>' or '<'
  var $varname;  // var name
  var $value;    // integer
};
class SceneData
{
  var $description; // Text
  var $options;     // Array(scenename, option text)
  var $followup;    // scenename or null
  var $vars_set;
  
  function __construct()
  {
    $this->description = '';
    $this->options  = Array();
    $this->vars_set = Array();
    $this->followup = null;
  }
};

function StringifyCondList($condlist)
{
  foreach($condlist as &$cond)
    $cond = "{$cond->varname}{$cond->type}{$cond->value}";
  return join(',', $condlist);
}

class Scene
{
  var $name;
  var $fixed_data;         // SceneData
  var $cond_data;          // Array( Array(Condition) , SceneData)
  
  function __construct()
  {
    $this->fixed_data = new SceneData;
    $this->cond_data = Array();
  }
};
class World
{
  var $scenes; // Array(scenename => Scene)
};

function ParseStory($description)
{
  $world      = new World;
  $scene      = null;
  $cond_stack = Array();
  $data_stack = Array();
  $data       = null;
  $cur_cond   = -1;
  $order = 0;
  foreach(explode("\n", $description) as $line)
  {
    $line = trim($line);
    if($line == '') continue;
    
    if($line[0] == '[')
    {
      #print_r($scene);
      $scenename = preg_replace('/[^a-zA-Z0-9_]/', '', $line);
      $scene        = &$world->scenes[$scenename];
      $scene = new Scene;
      $scene->name  = $scenename;
      $cond_stack   = Array();
      $data         = &$scene->fixed_data;
      $cur_cond     = -1;
      continue;
    }
    if($line[0] == '{') continue;
    if(preg_match('/^([=<>]) +([A-Z_0-9]+) +([0-9 ]+)$/', $line, $mat))
    {
      unset($data);
      $cond = new Condition;
      $cond->type    = $mat[1];
      $cond->varname = $mat[2];
      $cond->value   = $mat[3];
      $data_stack[] = $cur_cond;
      $cond_stack[] = $cond;
      $scene->cond_data[] = Array($cond_stack, new SceneData);
      $n = count($scene->cond_data) - 1;
      $cur_cond = $n;
      #print "[in scene {$scene->name}] Entering ... condition $n (".StringifyCondList($cond_stack).")\n";
      $data = &$scene->cond_data[$n][1];
      continue;
    }
    if($line[0] == '}')
    {
      unset($data);
      array_pop($cond_stack); 
      $n = array_pop($data_stack);
      $cur_cond = $n;
      
      #print "[in scene {$scene->name}] Returning to condition $n (".StringifyCondList($cond_stack).")\n";
      
      if($n < 0)
        $data = &$scene->fixed_data;
      else
        $data = &$scene->cond_data[ $n ][1];
      continue;
    }
    if(preg_match('/^[$] +([A-Z_0-9]+) +([0-9]+)$/', $line, $mat))
    {
      #print "adding setvar {$mat[2]}\n";
      $data->vars_set[$mat[1]] = $mat[2];
      continue;
    }
    if(preg_match('/^: +([a-zA-Z_0-9]+) +(.*)$/', $line, $mat))
    {
      #print "adding opt {$mat[2]}\n";
      $data->options[] = Array($mat[1], sprintf('%05d:', $order++).$mat[2]."\n");
      continue;
    }
    if(preg_match('/^> +([a-zA-Z_0-9]+)$/', $line, $mat))
    {
      #print "adding opt {$mat[2]}\n";
      $data->followup = $mat[1];
      continue;
    }
    #print "adding text {$line}\n";
    $data->description .= sprintf('%05d:', $order++)."$line\n";
  }
  return $world;
}

function ParseAttr($attrfile)
{
  $result = Array();
  foreach(explode("\n", $attrfile) as $line)
  {
    if(preg_match('/^([A-Z_]+) *([0-9]+)$/', $line, $mat))
      $result[$mat[1]] = (int)$mat[2];
  }
  ksort($result);
  return $result;
}

function MatchCondition($condition, $attr)
{
  switch($condition->type)
  {
    case '=':
      foreach(explode(' ', $condition->value) as $option)
        if($attr[$condition->varname] == $option)
          return true;
      return false;
    case '>':
      return $attr[$condition->varname] > $condition->value;
    case '<':
      return $attr[$condition->varname] < $condition->value;
  }
  print "WHAT?\n";
  return true;
}
function MatchConditionList($conditionlist, $attr)
{
  foreach($conditionlist as $condition)
  {
    if(!MatchCondition($condition, $attr))
    {
      #print "No match:\n";
      #print_r($condition);
      #print_r($attr);
      return false;
    }
  }
  return true;
}
function RenderText($text, $attr)
{
  $result = '';
  $lines = explode("\n", $text);
  sort($lines);
  foreach($lines as $line)
  {
    $line = preg_replace('/^[0-9]+:/', '', $line);
    if(!empty($line) && $line[0] == '\\')
    {
      if($result[strlen($result)-1] == "\n")
        $result = substr($result, 0, strlen($result)-1);
      $result .= substr($line, 1)."\n";
    }
    else
      $result .= "$line\n";
  }
  return trim($result);
}
function RenderDescription($scene, $attr)
{
  $text = $scene->fixed_data->description;
  foreach($scene->cond_data as $cond_scenedata)
  {
    if(MatchConditionList($cond_scenedata[0], $attr))
      $text .= $cond_scenedata[1]->description;
  }
  return trim(RenderText($text, $attr));
}
function RenderOptions($scene, $attr, &$options)
{
  foreach($scene->fixed_data->options as $opt)
    $options[] = $opt;
  foreach($scene->cond_data as $cond_scenedata)
  {
    if(MatchConditionList($cond_scenedata[0], $attr))
      foreach($cond_scenedata[1]->options as $opt)
        $options[] = $opt;
  }
  foreach($options as &$option)
    $option[1] = RenderText($option[1], $attr);
}
function GetFollowup($scene, $attr, &$followup)
{
  if($scene->fixed_data->followup)
    $followup = $scene->fixed_data->followup;
  foreach($scene->cond_data as $cond_scenedata)
  {
    if(MatchConditionList($cond_scenedata[0], $attr)
    && $cond_scenedata[1]->followup)
      $followup = $cond_scenedata[1]->followup;
  }
}
function MogrifyAttrs($scene, &$attr)
{
  $options = $scene->fixed_data->vars_set;
  foreach($scene->cond_data as $cond_scenedata)
  {
    if(MatchConditionList($cond_scenedata[0], $attr))
      foreach($cond_scenedata[1]->vars_set as $k=>$opt)
        $options[$k] = $opt;
  }
  foreach($options as $k=>$v)
    $attr[$k] = $v;
}

$seen_storypoints = Array();
$story_everything = Array();

function WalkStory($world, $attr, $scenename)
{
  global $seen_storypoints, $story_everything;
  
  $scene = &$world->scenes[$scenename];
  if(!isset($scene))
  {
    print "Error: Scene $scenename missing\n";
    return;
  }
  #print_r($scene);
  
  $scene_key = $attr;
  $scene_key['CUR_ROOM'] = $scenename;
  $scene_key = json_encode($scene_key);
  $scene_key = sprintf('%2$s_%1$s', $scenename, sha1($scene_key));
  
  if(isset($seen_storypoints[$scene_key]))
    return;
  $seen_storypoints[$scene_key] = true;
  
  $description = '';
  $options     = Array();
  $next_scenes = Array();
  $newoptions = Array();
  
  for(;;)
  {
    $description .= RenderDescription($scene, $attr);
    RenderOptions($scene, $attr, $options);

    $followup = '';
    GetFollowup($scene, $attr, $followup);
    
    MogrifyAttrs($scene, $attr);

    foreach($options as &$option)
    {
      $targetscene = $option[0];
      $next_scenes[$targetscene] = $targetscene;
      
      $targetscene_key = $attr;
      $targetscene_key['CUR_ROOM'] = $targetscene;
      $targetscene_key = json_encode($targetscene_key);
      $targetscene_key = sprintf('%2$s_%1$s', $targetscene, sha1($targetscene_key));
      $newoptions[$option[1]] = $targetscene_key;
    }
    
    if($attr['GAMEOVER'])
      $description .= "\nGAME OVER\n";
  
    if(!$followup)
    {
      $description = trim($description);
      if($description != '')
        $description .= "\n";
      break;
    }
    $scene = &$world->scenes[$followup];
    if(!isset($scene))
    {
      print "Error: Scene $followup missing\n";
      break;
    }
    $description = trim($description);
    if($description != '')
      $description .= "\n\n";
  }
  
  $scenedesc = Array(
    'description' => $description,
    'options'     => $newoptions);
  
  $story_everything[$scene_key] = $scenedesc;

  foreach($next_scenes as $nextscene)
  {
    WalkStory($world, $attr, $nextscene);
  }
  
  #OptimizeStory();
}

function DumpStoryForPostprocessing()
{
  global $story_everything;
  foreach($story_everything as $k => $d)
  {
    print sprintf('%-6d', strlen($k));
    print $k;
    print sprintf('%-6d', strlen($d['description']));
    print $d['description'];
    print sprintf('%-6d', count($d['options']));
    foreach($d['options'] as $opt => $k)
    {
      print sprintf('%-6d', strlen($k));
      print $k;
      print sprintf('%-6d', strlen($opt));
      print $opt;
    }
  }
  #print json_encode($story_everything);
  return;
}

function phpstr($s)
{
  return '"'.
    str_replace(Array('\\', '"', "\n"),
                Array('\\\\', '\\"', '\\n'),
                $s).'"';
}
function phpcond($cond)
{
  $operator = $cond->type == '=' ? '==' : $cond->type;
  $conds = Array();
  foreach(explode(' ', $cond->value) as $v)
    $conds[] = "\$GLOBALS['{$cond->varname}'] $operator $v";
  return join(' || ', $conds);
}
function PHPstatename($statename)
{
  static $table;
  if(isset($table[$statename])) return $table[$statename];
  return $table[$statename] = count($table);
}
function SetIfMax(&$var, $value)
{
  if(!isset($var) || $value > $var) $var = $value;
}
function ConvertIntoPHP($world, $attr, $filename)
{
  $max_values = Array();
  foreach($attr as $k => $v)
    SetIfMax($max_values[$k], $v);
  foreach($world->scenes as $k => $d)
  {
    SetIfMax($max_values['CUR_ROOM'], PHPstatename($k));
    foreach($d->fixed_data->vars_set as $k => $v)
      SetIfMax($max_values[$k], $v);
    foreach($d->cond_data as $conditions)
      foreach($conditions[1]->vars_set as $k => $v)
        SetIfMax($max_values[$k], $v);
  }
  foreach($max_values as $k=>$v)
    if($v == 0)
    {
      unset($attr[$k]);
      unset($max_values[$k]);
    }

  $fp = fopen($filename, 'w');
  if(!$fp) return;
  
  fwrite($fp, "<?php\n");
  ob_start();
?>
define('MAGIC', 'StoryBook');
/* ATTRIBUTES */
$attrlist = Array(
  'CUR_ROOM' => Array(<?php print PHPstatename('BEGIN');
     ?>, <?php print $max_values['CUR_ROOM']; ?>)<?php
  foreach($attr as $k=>$v)
    echo ",\n  '$k' => Array($v, ", $max_values[$k], ")";
  print "\n);\n";
?>
/* INITIAL ATTRIBUTES */
foreach($attrlist as $k => $v)
  $GLOBALS[$k] = $v[0];

define('MAX_INTEGER', 0x1FFFFFFF);
/* Request handling */
function EncodeRequest()
{
  global $attrlist;
  $result = '';
  $value  = 0;
  $factor = 1;
  $remain = count($attrlist);
  foreach($attrlist as $attrname => $attrspecs)
  {
    $max      = $attrspecs[1];
    $flush = (--$remain == 0 || ($max*$factor >= MAX_INTEGER));
    if($flush)
    {
      $ndigits = ceil(log($factor) / log(36));
      for($base = 36; $base > 2; --$base)
      {
        $ndigits2 = ceil(log($factor) / log($base-1));
        if($ndigits2 != $ndigits) break;
        $ndigits = $ndigits2;
      }
      $item = base_convert($value, 10, $base);
      while(strlen($item) < $ndigits) $item = "0$item";
      $result .= $item;
      $factor  = 1;
      $value   = 0;
    }
    $value += ($GLOBALS[$attrname] ^ $attrspecs[0]) * $factor;
    $factor *= ($max+1);
  }
  $cksum = sprintf('%s', substr(base_convert(sha1(MAGIC.$result),16,36),0,1));
  $result .= $cksum;
  return $result;
}
function DecodeRequest()
{
  if(!isset($_GET['req'])) return true;
  global $attrlist, $max;
  $req = $_GET['req'];
  $pos = 0;

  $value  = 0;
  $factor = 1;
  $remain = count($attrlist);
  $do_list = Array();
  foreach($attrlist as $attrname => $attrspecs)
  {
    $max      = $attrspecs[1];
    $flush = (--$remain == 0 || ($max*$factor >= MAX_INTEGER));
    if($flush)
    {
      $ndigits = ceil(log($factor) / log(36));
      for($base = 36; $base > 2; --$base)
      {
        $ndigits2 = ceil(log($factor) / log($base-1));
        if($ndigits2 != $ndigits) break;
        $ndigits = $ndigits2;
      }
      if($pos + $ndigits > strlen($req)) return false;
      $item = substr($req, $pos, $ndigits);
      $pos += $ndigits;
      $value  = base_convert($item, $base, 10);
      foreach($do_list as $do_attrname)
      {
        $do_attrspecs = $attrlist[$do_attrname];
        $do_max       = $do_attrspecs[1];
        $val = ($value % ($do_max+1));
        $val ^= $do_attrspecs[0];
        #print "sets $do_attrname to $val<br>";
        $GLOBALS[$do_attrname] = $val;
        $value = (int)($value / ($do_max+1));
      }
      if($value != 0) return false;
      $do_list = Array();
      $factor  = 1;
    }
    $do_list[] = $attrname;
    $factor *= ($max+1);
  }

  $result = substr($req,0,$pos);
  $cksum = sprintf('%s', substr(base_convert(sha1(MAGIC.$result),16,36),0,1));
  if(substr($req, $pos) != $cksum)
  {
    return false;
  }
  return true;
}
if(!DecodeRequest()) die('Request error');
$DESC    = '';
$OPTIONS = Array();

function AppendDesc(&$my_desc)
{
  $lines = explode("\n", $my_desc);
  sort($lines);
  global $DESC;
  $DESC .= join("\n", $lines)."\n";
  $my_desc = '';
}
function Collect($state)
{
  global $OPTIONS;
  $next = $state;
  do {
  $assigns = Array();
  $my_desc = '';
  $state   = $next;
  $next    = -1;
  switch($state)
  {
<?php
  foreach($world->scenes as $k => $d)
  {
    echo "    case ", PHPstatename($k), ": // $k:\n";
    echo "      \$my_desc .= ".phpstr($d->fixed_data->description).";\n";
    foreach($d->fixed_data->options as $option)
    echo "      \$OPTIONS[", phpstr($option[1]), '] = ', PHPstatename($option[0]), ";\n";
    foreach($d->fixed_data->vars_set as $vname => $vvalue)
    echo "      \$assigns['$vname'] = $vvalue;\n";
   
    foreach($d->cond_data as $conditions)
    {
      $scene      = $conditions[1];
      $conditions = $conditions[0];
      foreach($conditions as &$condition)
        $condition = phpcond($condition);
      echo "      if((", join(') && (', $conditions), ")) {\n";
      if($scene->description != '')
        echo "        \$my_desc .= ".phpstr($scene->description).";\n";
      foreach($scene->options as $option)
      echo "        \$OPTIONS[", phpstr($option[1]), '] = ', PHPstatename($option[0]), ";\n";
      foreach($scene->vars_set as $vname => $vvalue)
      echo "        \$assigns['$vname'] = $vvalue;\n";
      if($scene->followup)
      {
      echo "        \$next = ", PHPstatename($scene->followup), "; /* ", $scene->followup, " */\n";
      }
      echo "      }\n";
    }
    if($d->fixed_data->followup)
    echo "      \$next = ", PHPstatename($d->fixed_data->followup), "; /* ", $d->fixed_data->followup, " */\n";
    echo "      break;\n";
  }
?>
  }
  AppendDesc($my_desc);
  foreach($assigns as $k => $v)
    $GLOBALS[$k] = $v;
  if($GLOBALS['GAMEOVER'])
  {
    global $DESC;
    $DESC .= "99998:\n99999:GAME OVER\n";
  }
  } while($next >= 0);
}
$req = EncodeRequest();
Collect($CUR_ROOM);
<? print '?'.'>'; ?>
<html><head>
 <title>Game title here</title>
 <link rel="stylesheet" href="game.css"></head><body>
<? print '<'.'?php'; ?>
 echo '<h2>', 'Page ', preg_replace('/,.*/', '', $req), '</h2>', "\n";
 $desc_lines = explode("\n", preg_replace("/\n$/", '', $DESC));
 $nl = '';
 foreach($desc_lines as $line)
 {
   if($line == '') continue;
   $line = preg_replace('/^[0-9]*:/', '', $line);
   if($line != '' && $line[0] == '\\')
     $line = substr($line, 1);
   else
     echo $nl;
   echo htmlspecialchars($line);
   $nl = '&nbsp;<br>';
 }
<? print '?'.'>'; ?>
 <p><ul>
<? print '<'.'?php'; ?>
 ksort($OPTIONS);
 foreach($OPTIONS as $label => $target)
 {
   $CUR_ROOM = $target;
   $link = '?req='.(EncodeRequest());
   echo '<li>';
   $label_html = htmlspecialchars(preg_replace('/^[0-9]*:/', '', $label));
   echo '<a href="',htmlspecialchars($link),'">',$label_html,'</a>';
   echo '</li>';
 }
 if(empty($OPTIONS))
 {
   echo '<li><a href="?">Start a new game</a></li>';
 }
<? print '?'.'>'; ?>
 </ul>
 <hr><small>This is part of The Author's The Game interactive book</small></body></html>
<?php
  fwrite($fp, ob_get_clean());
  fclose($fp);
}

# function OptimizeStory()
# {
#   global $story_everything;
#   for(;;)
#   {
#     print count($story_everything)." storypoints\n";
#     $hashes  = Array();
#     $combining_hash   = null;
#     $combine          = Array();
#     $combine_preserve = null;
#     foreach($story_everything as $scene_key => $scene_desc)
#     {
#       $hash = sha1(serialize($scene_desc));
#       if(!isset($combining_hash))
#       {
#         if(!array_key_exists($hash, $hashes))
#           $hashes[$hash] = $scene_key;
#         else
#         {
#           // Duplicate!
#           $combine_preserve = $hashes[$hash];
#           $combine = Array($scene_key => true);
#           $combining_hash = $hash;
#         }
#       }
#       else if($hash == $combining_hash)
#         $combine[$scene_key] = true;
#     }
#     if(isset($combining_hash))
#     {
#       print "Merging storypoints\n";
#       foreach($story_everything as $scene_key => &$scene_desc)
#       {
#         if(array_key_exists($scene_key, $combine))
#         {
#           unset($story_everything[$scene_key]);
#         }
#         else
#         {
#           foreach($scene_desc['options'] as $opttext => &$optlabel)
#             if(array_key_exists($optlabel, $combine))
#               $optlabel = $combine_preserve;
#           unset($optlabel);
#         }
#       }
#       unset($scene_desc);
#     }
#     else
#     {
#       break;
#     }
#   }
# }

$world = ParseStory(file_get_contents('story.dat'));
$attr  = ParseAttr(file_get_contents('attr.dat'));

#print_r($world); exit;
ConvertIntoPHP($world, $attr, 'story_result_php.php');

#print_r($world);
WalkStory($world, $attr, 'BEGIN');
#WalkStory($world, $attr, 'BEGIN_DOOR_SW1');

DumpStoryForPostprocessing();
#OptimizeStory();
#print_r($story_everything);
