#include <map>
#include <string>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <vector>
#include <algorithm>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

std::vector<std::string> InputTokens;
std::vector<std::string> OutputActions;
std::vector<std::string> StateNames;

struct State
{
    size_t action;
    std::vector<std::pair<size_t/*input*/,size_t/*state*/> > transitions;
    
    State() : action(),transitions() { }
};

class FindAddMap
{
    std::map<std::string, size_t> map;
    std::vector<std::string>& target;
public:
    FindAddMap(std::vector<std::string>& t) : map(), target(t) { }
    
    size_t Get(const std::string& s)
    {
        std::map<std::string, size_t>::iterator
            i = map.lower_bound(s);
        if(i == map.end() || i->first != s)
        {
            size_t n = target.size();
            target.push_back(s);
            map.insert(i, std::make_pair(s,n));
            return n;
        }
        return i->second;
    }
};

template<typename CharT>
const std::basic_string<CharT>
str_replace(const std::basic_string<CharT>& search,
            const std::basic_string<CharT>& with,
            const std::basic_string<CharT>& where)
{
    std::basic_string<CharT> result;
    for(std::string::size_type a=0; a < where.size(); )
    {
        std::string::size_type b = where.find(search, a);
        if(b == where.npos)
        {
            //result += where.substr(a);
            result.insert(result.end(), where.begin()+a, where.end());
            break;
        }
        //result += where.substr(a, b-a);
        result.insert(result.end(), where.begin()+a, where.begin()+b);
        result += with;
        a = b + search.size();
    }
    return result;
}

std::string str_replace(const char*a, const char*b, const std::string& c)
{
    return str_replace(std::string(a),std::string(b),c);
}

std::string htmlfix(const std::string& s)
{
    std::string result;
    for(size_t a=0; a<s.size(); ++a)
    {
        if(s[a] == '<') result += "&lt;";
        else if(s[a] == '&') result += "&amp;";
        else if(s[a] == '\n') result += "&nbsp;<br>";
        else result += s[a];
    }
    return result;
}

std::string pagename(unsigned long n)
{
    char Buf[512], *suffix = Buf+128;
    *suffix = 0;
    if(n != 1)
    {
        std::srand((n^15)+10);
        for(unsigned n=0; n<2; ++n)
            suffix[n] = 'a' + (char)(rand()%26);
        suffix[2] = 0;
    }
    sprintf(Buf, "page%lu%s.html", n, suffix);
    return Buf;
}

int main()
{
    setvbuf(stdout, 0, _IOLBF, 0);
    
    FindAddMap state_map(StateNames), output_map(OutputActions), input_map(InputTokens);
    
    // Mandatory option number 0
    input_map.Get("<eps>");
    output_map.Get("<eps>");
    // For transitioning to final state (also from BEGIN to actual begin)
    input_map.Get("DUMMYCHOICE");
    output_map.Get("DUMMYOUTPUT");

    state_map.Get("BEGIN"); // BEGIN is at 0
    state_map.Get("END");   // This puts FINAL state into state 1 (0 must be BEGIN)

    std::map<size_t, State> states;
    for(;;)
    {
        char Buf[65536];
        
        Buf[0]='\0';
        fread(Buf,1,6,stdin); Buf[6]='\0'; if(!Buf[0]) break;
        
        {int l=atoi(Buf); Buf[l]='\0'; fread(Buf,1,l,stdin); }
        const size_t src_state_id = state_map.Get(Buf);
        
        fread(Buf,1,6,stdin); Buf[6]='\0';
        {int l=atoi(Buf); Buf[l]='\0'; fread(Buf,1,l,stdin); }
        
        const size_t action = output_map.Get(Buf);
        
        State& sc = states[src_state_id];
        sc.action = action;
        
        fread(Buf,1,6,stdin); Buf[6]='\0';
        int n_options = atoi(Buf);
        for(int n=0; n<n_options; ++n)
        {
            fread(Buf,1,6,stdin); Buf[6]='\0';
            {int l=atoi(Buf); Buf[l]='\0'; fread(Buf,1,l,stdin); }
            const size_t dest_state_id = state_map.Get(Buf);

            fread(Buf,1,6,stdin); Buf[6]='\0';
            {int l=atoi(Buf); Buf[l]='\0'; fread(Buf,1,l,stdin); }
            const size_t text_id = input_map.Get(Buf);
            
            sc.transitions.push_back(std::pair<size_t,size_t> (text_id,dest_state_id));
        }
    }

#if 1
    int pip[2]; pipe(pip);
    if(!fork())
    {
        dup2(pip[1], 1); close(pip[0]); close(pip[1]);
        
        printf("%u %u %u %lu\n",
            0, /* openfst's begin */
            2, /* actual begin */
            1, /* dummy input token */
            states[2].action); /* implicit output at beginning */
        printf("%u\n",
            1); /* end state */

        for(std::map<size_t, State>::const_iterator
            i = states.begin(); i != states.end(); ++i)
        {
            if(i->second.transitions.empty()) /* Transition into the final state */
            {
                printf("%lu %u %u %lu\n",
                    i->first, /* source state */
                    1, /* end state */
                    1, /* dummy input token */
                    i->second.action); /* dummy output token */
            }
            else for(size_t n=0; n<i->second.transitions.size(); ++n)
            {
                const size_t input_token  = i->second.transitions[n].first;
                const size_t target_state = i->second.transitions[n].second;
                printf("%lu %lu %lu %lu\n",
                    i->first,                     /* source state */
                    target_state,                 /* target state */
                    input_token,                  /* input token */
                    i->second.action/*
                    states[target_state].action*/); /* output token */
            }
        }
        fflush(stdout);
        _exit(0);
    }
    close(pip[1]);
    states.clear();
    
    static const char tempfn[] = "bin.fst";
    if(!fork())
    {
        dup2(pip[0], 0); close(pip[0]);
        int outfd = open(tempfn, O_WRONLY | O_CREAT, 0644);
        dup2(outfd, 1); close(outfd);
        execl("openfst/src/bin/fstcompile", "fstcompile", NULL);
        //execlp("tee", "tee", "test.fst", NULL);
    }
    close(pip[0]);
    while(wait(0) >= 0) {}
    
    pipe(pip);
    if(!fork())
    {
        close(pip[0]);
        dup2(pip[1], 1); close(pip[1]);
        execl("openfst/src/bin/fstminimize", "fstminimize", tempfn, NULL);
        //execl("/bin/cat", "cat", tempfn, NULL);
    }
    close(pip[1]);
    int read_minimized = pip[0];
    pipe(pip);
    if(!fork())
    {
        dup2(read_minimized, 0); close(read_minimized);
        close(pip[0]);
        dup2(pip[1], 1); close(pip[1]);
        execl("openfst/src/bin/fstprint", "fstprint", NULL);
    }
    close(pip[1]);
    FILE* result = fdopen(pip[0], "rt");
    
    for(;;)
    {
        char Buf[65536];
        if(!fgets(Buf,sizeof Buf, result)) break;
        int istate, ostate, isym, osym;
        int n = sscanf(Buf, "%d %d %d %d", &istate,&ostate,&isym,&osym);
        
        if(n == 4)
        {
            states[istate].action = osym;
            states[istate].transitions.push_back(std::pair<size_t,size_t>(isym,ostate));
        }
        else if(n == 1) // final state
        {
            // ignore
        }
    }
    fclose(result);
    while(wait(0) >= 0) {}
#endif

printf(
"{\\rtf1\\ansi\\ansicpg1252\\deff0\\deflang1035\\deflangfe1041\\deftab1304{\\fonttbl{\\f0\\froman\\fprq2\\fcharset0 Cambria;}{\\f1\\froman\\fprq2\\fcharset0 Century Schoolbook;}}\r\n"
"{\\colortbl ;\\red79\\green129\\blue189;\\red0\\green0\\blue255;}\r\n"
"{\\stylesheet{ Normal;}{\\s1 heading 1;}{\\s2 heading 2;}}\r\n"
"{\\*\\generator Msftedit 5.41.15.1515;}\\viewkind4\\uc1\r\n"
);

    size_t expected_page = 1;
    for(std::map<size_t, State>::iterator
        i = states.begin(); i != states.end(); ++i)
    {
        if(i->first == 0          /* Ignore dummy state */
        || i->second.action == 0) /* Ignore <eps> */
        {
            continue;
        }
        
        while(expected_page < i->first)
        {
            printf("%s\\par\r\n"
                   "\\page\r\n",
                "This page intentionally left blank");
            ++expected_page;
        }
        
        printf(
"\\pard\\keep\\keepn\\nowidctlpar\\s2\\sb200\\sl276\\slmult1\\cf1\\b\\f0\\fs26 PAGE %lu\\par\r\n"
"\\pard\\nowidctlpar\\cf0\\b0\\f1\\fs22\\par\r\n"
    , i->first
);
    
        FILE* fp = fopen(pagename(i->first).c_str(), "wt");
        
        fprintf(fp,
            "<html><head><title>Game title here</title>"
            "<link rel=\"stylesheet\" href=\"game.css\">"
            "</head>"
            "<body><h2>Page %lu</h2>", i->first);
        
        std::sort(i->second.transitions.begin(),
                  i->second.transitions.end());
        
        //printf("[PAGE %lu]\n", i->first);
        
        printf("%s\n",
            str_replace("\n", "\\par\r\n",
                OutputActions[i->second.action]
            ).c_str());
        
        fprintf(fp,
            htmlfix(
                OutputActions[i->second.action]
            ).c_str());
        
        fprintf(fp, "<ul>");
        
        for(size_t n=0; n<i->second.transitions.size(); ++n)
        {
            if(i->second.transitions[n].first == 1)
            {
                printf(
"\\pard Page {\\field{\\*\\fldinst{HYPERLINK \"\\\\l \"_PAGE_%lu\"\"}}{\\fldrslt{\\cf2\\ul %lu}}}\\cf0\\ulnone\\f1\\fs22 : %s\\par\r\n",
                    1lu, 1lu,
                    "Start a new game");
                //printf(">1>%s\n",
                //    "Start a new game");
                
                fprintf(fp,
                    "<li><a href=\"%s\">Start a new game</a></li>",
                    pagename(1lu).c_str());
                
                continue; /* Ignore DUMMYCHOICEs */
            }

            const std::string& label = InputTokens[i->second.transitions[n].first];
            const long tgt = i->second.transitions[n].second;
            
            /*printf(">%lu>%s\n",
                tgt,
                label.c_str()
            );
            */
            printf(
"\\pard Page {\\field{\\*\\fldinst{HYPERLINK \"\\\\l \"_PAGE_%lu\"\"}}{\\fldrslt{\\cf2\\ul %lu}}}\\cf0\\ulnone\\f1\\fs22 : %s\\par\r\n",
                tgt, tgt,
                label.c_str()
            );
            
            fprintf(fp,
                "<li><a href=\"%s\">%s</a></li>",
                    pagename(tgt).c_str(),
                    htmlfix(label).c_str());
        }
        printf("\\page\r\n");
        ++expected_page;
        
        fprintf(fp, "</ul>"
            "<hr>"
            "<small>This is part of The Author's The Game interactive book</small>"
            "</body></html>");
        
        fclose(fp);
    }
    printf(
"\\pard\\nowidctlpar\\sa200\\sl276\\slmult1\\par\r\n}\r\n"
);
}
